import csv
import logging

from opennem.core.loader import load_data
from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility
from opennem.importer.mms import mms_import

logger = logging.getLogger(__name__)


def import_mms_emissions() -> None:
    mms = mms_import()

    facility_poll_map = {
        "CALL_A_4": 0.93134349,
        "CALL_B_1": 0.93134349,
        "CALL_B_2": 0.93134349,
        "SWAN_E": 0.42795234,
        "OSB-AG": 0.544686,
    }

    for station in mms:
        for facility in station.facilities:
            if facility.emissions_factor_co2 and not facility.code.endswith(
                "NL1"
            ):
                facility_poll_map[facility.code] = {
                    "intensity": facility.emissions_factor_co2,
                    "name": station.name,
                    "fueltech": facility.fueltech.code
                    if facility.fueltech
                    else "",
                }

    session = SessionLocal()

    for fac_code, pol_value in facility_poll_map.items():
        facility = (
            session.query(Facility)
            .filter_by(code=fac_code)
            .filter_by(network_id="NEM")
            .one_or_none()
        )

        if not facility:
            logger.info(
                "could not find facility: {} {}".format(fac_code, pol_value)
            )

            continue

        facility.emissions_factor_co2 = clean_float(pol_value)
        session.add(facility)

    session.commit()

    return None


def import_emissions_csv() -> None:
    EMISSION_MAPS = [
        {"filename": "wem_emissions.csv", "network": "WEM"},
        {"filename": "nem_emissions.csv", "network": "NEM"},
    ]

    for m in EMISSION_MAPS:
        import_emissions_map(m["filename"], m["network"])


def import_emissions_map(file_name: str, network_id: str) -> None:
    session = SessionLocal()

    content = load_data(file_name, from_project=True, skip_loaders=True)

    csv_content = content.splitlines()
    csvreader = csv.DictReader(csv_content)

    for rec in csvreader:
        facility_code = normalize_duid(rec["facility_code"])
        emissions_intensity = clean_float(rec["emissions_factor_co2"])

        if not facility_code:
            logger.info("No emissions intensity for {}".format(facility_code))
            continue

        facility = (
            session.query(Facility)
            .filter_by(code=facility_code)
            .filter_by(network_id=network_id)
            .one_or_none()
        )

        if not facility:
            logger.info("No stored facility for {}".format(facility_code))
            continue

        facility.emissions_factor_co2 = emissions_intensity
        session.add(facility)
        logger.info(
            "Updated {} to {}".format(facility_code, emissions_intensity)
        )

    session.commit()

    return None


if __name__ == "__main__":
    import_emissions_csv()



from textwrap import dedent
from typing import List, Optional, Union

from opennem.api.stats.schema import ScadaDateRange
from opennem.core.normalizers import normalize_duid
from opennem.schema.network import NetworkNEM, NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod


def station_id_case(station_codes: List[str]) -> str:
    return ",".join(
        ["'{}'".format(i) for i in map(normalize_duid, station_codes)]
    )


def observation_query(
    station_codes: List[str],
    interval: TimeInterval,
    network: NetworkSchema = NetworkNEM,
    period: Optional[TimePeriod] = None,
    scada_range: Optional[ScadaDateRange] = None,
    year: Optional[Union[str, int]] = None,
) -> str:
    if isinstance(year, int):
        year = str(year)

    timezone = network.timezone_database

    if not timezone:
        timezone = "AEST"

    __query = """
    select
        t.observation_time at time zone '{timezone}',
        t.station_id,
        avg(t.temp_air) as temp_avg,
        min(t.temp_air) as temp_min,
        max(t.temp_air) as temp_max
    from
        (
        select
            time_bucket_gapfill('{trunc}', observation_time) as observation_time,
            fs.station_id as station_id,
            interpolate(avg(fs.temp_air)) as temp_air
        from bom_observation fs
        where
            fs.station_id in ({station_codes})
            and fs.observation_time <= {date_end}
            and fs.observation_time >= {date_start_condition}
        group by 1, 2
        )
    as t
    group by 1, 2
    order by 1 desc, 2 desc
    """

    date_end = "now()"

    if scada_range:
        date_end = scada_range.get_end_sql()

    date_start_condition = ""

    if period:
        date_start_condition = (
            "{date_end}::timestamp - '{period}'::interval".format(
                date_end=date_end, period=period.period_sql
            )
        )

    if year:
        date_start_condition = "'{year}-01-01'::date".format(year=year)

    if not period and not year:
        if not scada_range:
            raise Exception("require a scada range ")

        date_start_condition = "'{}'".format(scada_range.get_start())

    query = dedent(
        __query.format(
            station_codes=station_id_case(station_codes),
            trunc=interval.interval_sql,
            timezone=timezone,
            date_start_condition=date_start_condition,
            date_end=date_end,
        )
    )

    return query


def observation_query_all(
    station_codes: List[str],
    scada_range: ScadaDateRange,
    network: NetworkSchema = NetworkNEM,
) -> str:

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """
        select
            date_trunc('month', t.observation_time at time zone '{timezone}') as observation_month,
            t.station_id,
            avg(t.temp_avg),
            min(t.temp_min),
            max(t.temp_max)
        from
            (
                select
                    time_bucket_gapfill('1 day', observation_time) as observation_time,
                    fs.station_id,
                    interpolate(avg(fs.temp_air)) as temp_avg,
                    interpolate(min(fs.temp_air)) as temp_min,
                    interpolate(max(fs.temp_air)) as temp_max
                from bom_observation fs
                where
                    fs.station_id in ({station_codes})
                    and fs.observation_time <= {date_end}
                    and fs.observation_time >= {date_start}
                group by 1, 2
            ) as t
        group by 1, 2
        order by 1 desc, 2 desc
    """

    date_end = scada_range.get_end_sql()
    date_start = scada_range.get_start_sql()

    query = __query.format(
        station_codes=station_id_case(station_codes),
        timezone=timezone,
        date_start=date_start,
        date_end=date_end,
    )

    return query



import logging

import click
from scrapy.utils.python import garbage_collect

from opennem.db.load_fixtures import load_bom_stations_json, load_fixtures
from opennem.diff.facility_diff import run_diff
from opennem.importer.all import run_all
from opennem.importer.db import init as db_init
from opennem.importer.mms import mms_export
from opennem.importer.opennem import opennem_export, opennem_import
from opennem.settings import settings

logger = logging.getLogger("opennem.cli")


@click.group()
def main():
    pass


@click.command()
def crawl():
    logger.info("crawl @TODO")


@click.command()
def diff():
    run_diff()


@click.group()
def cmd_db():
    pass


@click.command()
def cmd_db_init():
    db_init()


@click.command()
def cmd_db_fixturer():
    load_fixtures()


@click.group()
def cmd_import():
    pass


@click.group()
def cmd_export():
    pass


@click.command()
def cmd_import_opennem():
    opennem_import()


@click.command()
def cmd_import_mms():
    mms_export()


@click.command()
def cmd_import_all():
    run_all()


@click.command()
def cmd_export_opennem():
    opennem_export()


@click.command()
def cmd_export_all():
    run_all()


@click.group()
def cmd_weather():
    pass


@click.command()
def cmd_weather_init():
    load_bom_stations_json()


main.add_command(crawl)
main.add_command(diff)
main.add_command(cmd_db, name="db")
main.add_command(cmd_import, name="import")
main.add_command(cmd_export, name="export")
main.add_command(cmd_weather, name="weather")

cmd_import.add_command(cmd_import_opennem, name="opennem")
cmd_import.add_command(cmd_import_mms, name="mms")
cmd_import.add_command(cmd_import_all, name="all")

cmd_export.add_command(cmd_export_opennem, name="opennem")
cmd_export.add_command(cmd_export_all, name="all")

cmd_db.add_command(cmd_db_init, name="init")
cmd_db.add_command(cmd_db_fixturer, name="load_fixtures")

cmd_weather.add_command(cmd_weather_init, name="init")

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.error("User interrupted")
    except Exception as e:
        logger.error(e)

        if settings.debug:
            import traceback

            traceback.print_exc()

    finally:
        garbage_collect()



import os
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlparse

BASE_ENV_FILE_NAME = ".env"


def replace_database_in_url(db_url, db_name):
    """
    replaces the database portion of a database connection URL with db_name

    @param db_name database name to replace with
    """

    db_url_parsed = urlparse(db_url)
    db_url_parsed = db_url_parsed._replace(path=f"/{db_name}")

    return db_url_parsed.geturl()


# cache cwd
_cwd = Path(os.getcwd())


def _is_env_file(env_file_path: str) -> bool:
    """
    Checks if env file exists in current working directory
    """
    return (_cwd / Path(env_file_path)).is_file()


def load_env_file(environment: Optional[str]) -> List[str]:
    """
    Returns a list of env files to load
    """
    env_files = [BASE_ENV_FILE_NAME]

    if environment:
        environment_suffix = environment.strip().lower()

        env_files.append(
            "{}.{}".format(BASE_ENV_FILE_NAME, environment_suffix)
        )

    env_files_exist = [str(i) for i in env_files if _is_env_file(i)]

    return env_files_exist



from typing import Dict, List, Optional, Union

from geojson_pydantic.features import Feature, FeatureCollection
from geojson_pydantic.geometries import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from pydantic import BaseModel

Geometry = Union[
    Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon
]


class FacilityGeoBase(BaseModel):
    class Config:
        arbitrary_types_allowed = True


class FacilityGeometryCollection(FacilityGeoBase, GeometryCollection):
    geometries: Optional[List[Geometry]]


class FacilityFeature(Feature):
    """
        Update to default to GeometryCollection
    """

    geometry: Optional[Geometry]


class FacilityGeo(FacilityGeoBase, FeatureCollection):
    name: str = "opennem"
    crs: Optional[Dict]
    features: List[FacilityFeature]




"""
    Set of stations


"""
from collections import UserList
from typing import List, Optional

from opennem.exporter.encoders import opennem_serialize
from opennem.schema.opennem import StationSchema


def _str_comp(subject: str, value: str) -> bool:
    return subject.trim().lower() == value.trim().lower()


class StationSet(UserList):
    """

    """

    # def __init__(self,):
    #     pass

    def get(self, key: int) -> Optional[StationSchema]:
        _entry = list(filter(lambda s: s.id == key, self.data))

        if not _entry or len(_entry) == 0:
            return None

        return _entry.pop()

    def get_by(self, **kwargs):
        print(kwargs)

    def get_name(self, name: str) -> List[StationSchema]:
        _entries = list(filter(lambda s: _str_comp(s.name, name), self.data))

        return _entries

    def get_code(self, code: str) -> Optional[StationSchema]:
        _entries = list(filter(lambda s: s.code == code, self.data))

        if not _entries or len(_entries) == 0:
            return None

        if len(_entries) > 1:
            raise Exception("More than one station with a code")

        return _entries.pop()

    def one(self):
        pass

    def add(self, station: StationSchema):
        if not station.code:
            raise Exception("Require a station code")

        _key = station.code

        if self.get_code(_key):
            raise Exception("Duplicate code {}".format(_key))

        self.data.append(station)

        return self

    def add_dict(self, station_dict: dict):
        station = StationSchema(**station_dict)

        self.add(station)

    def as_list(self) -> List[StationSchema]:
        return self.data

    def json(self, indent=None):
        _data = [i.dict() for i in self.data]

        return opennem_serialize(_data, indent=indent)

    @property
    def length(self) -> int:
        return len(self.data)



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, validator

from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeIntervalAPI, TimePeriodAPI
from opennem.utils.numbers import sigfig_compact
from opennem.utils.timezone import get_current_timezone


class ScadaInterval(object):
    date: datetime
    value: Optional[float]

    def __init__(self, date: datetime, value: Optional[float] = None):
        self.date = date
        self.value = value


class ScadaReading(Tuple[datetime, Optional[float]]):
    pass


class UnitScadaReading(BaseModel):
    code: str
    data: List[ScadaReading]


class StationScadaReading(BaseModel):
    code: str
    facilities: Dict[str, UnitScadaReading]


class OpennemDataHistory(BaseConfig):
    start: Union[datetime, date]
    last: Union[datetime, date]
    interval: str
    data: List[Optional[float]]

    @validator("data")
    def validate_data(cls, data_value):
        data_value = list(
            map(lambda x: sigfig_compact(x) if x else None, data_value)
        )

        return data_value


class OpennemData(BaseConfig):
    id: Optional[str]
    type: Optional[str]
    region: Optional[str]
    fuel_tech: Optional[str]

    network: Optional[str]
    data_type: str
    code: str = ""
    units: str

    interval: TimeIntervalAPI
    period: Optional[TimePeriodAPI]

    history: OpennemDataHistory


class OpennemDataSet(BaseConfig):
    type: Optional[str]
    version: Optional[str]
    network: Optional[str]
    code: str
    region: Optional[str]
    created_at: Optional[datetime]

    data: List[OpennemData]

    def append_set(self, subject_set: Optional[OpennemDataSet] = None):
        if not subject_set:
            return None

        if not subject_set.data:
            return None

        if not isinstance(subject_set.data, list):
            return None

        if len(subject_set.data):
            self.data += subject_set.data

        return None


class DataQueryResult(BaseConfig):
    interval: datetime
    result: Union[float, int, None, Decimal]
    group_by: Optional[str]


class ScadaDateRange(BaseConfig):
    start: datetime
    end: datetime
    network: Optional[NetworkSchema]

    def _get_value_localized(self, field_name: str = "start"):
        timezone = get_current_timezone()
        date_aware = getattr(self, field_name)

        if self.network:
            timezone = self.network.get_timezone()

        date_aware = date_aware.astimezone(timezone)

        return date_aware

    def get_start_year(self):
        return self.start.year

    def get_start(self):
        return self._get_value_localized("start")

    def get_end(self):
        return self._get_value_localized("end")

    def get_start_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(
            self.get_start() if not as_date else self.get_start().date()
        )

    def get_end_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(
            self.get_end() if not as_date else self.get_end().date()
        )



import decimal
from math import floor, log, pow
from typing import Union

from opennem.settings import settings

__log10 = 2.302585092994046

DEFAULT_PRECISION = settings.precision_default

ctx = decimal.Context()
ctx.prec = 20


def float_to_str(f):
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    d1 = ctx.create_decimal(repr(f))
    return format(d1, "f")


def cast_number(number: any) -> float:
    """ Cast to a float """
    number_float = float(number)

    return number_float


def num_sigfigs(n, sig):
    """
    Adapted significant figure count

    """
    multi = pow(10, sig - floor(log(n) / __log10) - 1)
    return round(n * multi) / multi


def sigfig_compact(
    n: Union[float, int], precision: int = DEFAULT_PRECISION
) -> float:
    """
    Compact significant figure formatting
    """

    if not isinstance(n, float):
        n = cast_number(n)

    n_abs = abs(n)
    is_neg = n < 0

    if n == 0:
        return n

    if n_abs >= pow(10, precision):
        n = floor(n_abs)
    else:
        n = num_sigfigs(n_abs, precision)

    if is_neg:
        n *= -1

    return n


def human2bytes(s: str) -> int:
    """
    >>> human2bytes('1M')
    1048576
    >>> human2bytes('1G')
    1073741824
    """
    if s is None:
        return None
    try:
        return int(s)
    except ValueError:
        pass

    symbols = "BKMGTPEZY"
    letter = s[-1:].strip().upper()
    num = float(s[:-1])
    prefix = {symbols[0]: 1}

    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i + 1) * 10

    return int(num * prefix[letter])


def float_to_str(f: float) -> str:
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    d1 = ctx.create_decimal(repr(f))
    return format(d1, "f")



"""
    Utility functions for API exporter

"""
import json
import logging

from pydantic.main import BaseModel

from opennem.exporter.aws import write_to_s3
from opennem.exporter.local import write_to_local
from opennem.settings import settings

logger = logging.getLogger(__name__)


def write_output(
    path: str, stat_set: BaseModel, is_local: bool = False
) -> int:

    if settings.export_local:
        is_local = True

    if hasattr(stat_set, "json"):
        write_content = stat_set.json(exclude_unset=True)
    else:
        write_content = json.dumps(stat_set)

    byte_count = 0

    if is_local:
        byte_count = write_to_local(path, write_content)
    else:
        byte_count = write_to_s3(path, write_content)

    return byte_count



import json
import logging
from typing import List

from pydantic import ValidationError

from opennem.exporter.aws import write_to_s3
from opennem.exporter.csv import stations_csv_serialize
from opennem.exporter.geojson import stations_geojson_serialize
from opennem.exporter.local import write_to_local
from opennem.exporter.onjson import stations_json_serialize
from opennem.importer.all import run_all
from opennem.importer.opennem import opennem_import
from opennem.schema.opennem import StationSchema

logger = logging.getLogger("opennem.exporter")


def load_stations():
    with open("data/opennem.json") as fh:
        __data = json.load(fh)

    stations: List[StationSchema] = []

    for i in __data.values():
        try:
            stations.append(StationSchema(**i))
        except ValidationError as e:
            logger.error(
                "Error with record: {} {}: {}".format(i["code"], i["name"], e)
            )

    return stations


def stations_geojson_to_s3(stations: List[StationSchema]):
    stations_geojson = stations_geojson_serialize(stations)

    facilities_path = "geo/au_facilities.json"

    file_length = write_to_s3(facilities_path, stations_geojson)

    return file_length


def stations_geojson_to_local(stations: List[StationSchema]):
    stations_geojson = stations_geojson_serialize(stations)

    write_to_local("stations.geojson", stations_geojson)


def stations_csv_to_local(stations: List[StationSchema]):
    stations_csv = stations_csv_serialize(stations)

    write_to_local("stations.csv", stations_csv)


def stations_json_to_local(stations: List[StationSchema]):
    stations_json = stations_json_serialize(stations)

    write_to_local("stations.json", stations_json)


if __name__ == "__main__":
    run_all()
    stations = opennem_import()

    stations_geojson_to_s3(stations.copy())
    stations_geojson_to_local(stations.copy())
    stations_csv_to_local(stations.copy())
    stations_json_to_local(stations.copy())



from datetime import datetime, timedelta
from textwrap import dedent
from typing import List, Optional

from opennem.api.stats.controllers import get_scada_range, networks_to_in
from opennem.api.stats.schema import ScadaDateRange
from opennem.schema.network import NetworkSchema, NetworkWEM
from opennem.schema.time import TimeInterval, TimePeriod


def price_network_query(
    interval: TimeInterval,
    network: NetworkSchema = NetworkWEM,
    date_range: Optional[ScadaDateRange] = None,
    period: Optional[TimePeriod] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    if not date_range:
        date_range = get_scada_range(network=network, networks=networks_query)

    timezone = network.get_timezone(postgres_format=True)

    __query = """
        SET SESSION TIME ZONE '{timezone}';

        select
            time_bucket_gapfill('{trunc}', bs.trading_interval) AS trading_interval,
            bs.{group_field},
            coalesce(avg(bs.price), 0) as price
        from balancing_summary bs
        where
            bs.trading_interval <= '{date_max}' and
            bs.trading_interval >= '{date_min}' and
            {network_query}
            {network_region_query}
            1=1
        group by 1, 2
        order by 1 desc
    """

    group_field = "network_id"
    network_region_query = ""

    if network_region:
        network_region_query = f"bs.network_region='{network_region}' and "
        group_field = "network_region"

    network_query = "bs.network_id IN ({}) and ".format(
        networks_to_in(networks_query)
    )

    if not date_range:
        raise Exception("Require a date range")

    date_max = date_range.get_end()
    date_min = date_range.get_start()

    if period:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = dedent(
        __query.format(
            network_query=network_query,
            trunc=interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            group_field=group_field,
        )
    )

    return query


def power_network_fueltech_query(
    interval: TimeInterval,
    network: NetworkSchema = NetworkWEM,
    date_range: Optional[ScadaDateRange] = None,
    period: Optional[TimePeriod] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    if not date_range:
        date_range = get_scada_range(network=network, networks=networks_query)

    timezone = network.timezone_database

    __query = """
        select
            t.trading_interval at time zone '{timezone}',
            t.fueltech_code,
            sum(t.facility_power)
        from (
            select
                time_bucket_gapfill('{trunc}', fs.trading_interval) AS trading_interval,
                coalesce(
                    avg(fs.generated), 0
                ) as facility_power,
                fs.facility_code,
                ft.code as fueltech_code
            from facility_scada fs
            join facility f on fs.facility_code = f.code
            join fueltech ft on f.fueltech_id = ft.code
            where
                f.fueltech_id is not null and
                {network_query}
                {network_region_query}
                fs.trading_interval <= '{date_max}' and
                fs.trading_interval > '{date_min}'
                {fueltech_filter}
            group by 1, 3, 4
        ) as t
        group by 1, 2
        order by 1 desc
    """

    network_region_query = ""
    fueltech_filter = ""

    if network_region:
        network_region_query = f"f.network_region='{network_region}' and "
    else:
        fueltech_filter = "and f.fueltech_id not in ('imports', 'exports')"

    network_query = "fs.network_id IN ({}) and ".format(
        networks_to_in(networks_query)
    )

    if not date_range:
        raise Exception("require a date range")

    date_max = date_range.get_end()
    date_min = date_range.get_start()

    if period:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = dedent(
        __query.format(
            network_query=network_query,
            trunc=interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            fueltech_filter=fueltech_filter,
        )
    )

    return query


def energy_network_fueltech_query(
    network: NetworkSchema,
    year: Optional[int] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """
    Get Energy for a network or network + region
    based on a year
    """

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    __query = """
        select
            date_trunc('{trunc}', t.trading_interval at time zone '{timezone}') as trading_day,
            t.fueltech_id,
            sum(t.energy) / 1000 as fueltech_energy,
            sum(t.market_value) as fueltech_market_value,
            sum(t.emissions) as fueltech_emissions
        from
            (select
                time_bucket_gapfill('1 hour', fs.trading_interval) as trading_interval,
                f.fueltech_id,
                coalesce(
                    sum(eoi_quantity),
                    energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)),
                    0.0
                ) as energy,
                case when avg(bs.price) > 0 then
                    coalesce(
                        sum(eoi_quantity) * avg(bs.price),
                        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) * avg(bs.price),
                        0.0
                    )
                else 0.0
                end as market_value,
                case when avg(f.emissions_factor_co2) > 0 then
                    coalesce(
                        sum(eoi_quantity) * avg(f.emissions_factor_co2),
                        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) * avg(f.emissions_factor_co2),
                        0.0
                    )
                else 0.0
                end as emissions
            from facility_scada fs
                left join facility f on fs.facility_code = f.code
                left join balancing_summary bs on
                    bs.trading_interval = fs.trading_interval
                    and bs.network_id=fs.network_id
                    and bs.network_region = f.network_region
            where
                f.fueltech_id is not null and
                fs.trading_interval <= '{year_max}' and
                fs.trading_interval >= '{year_min}' and
                {network_query}
                {network_region_query}
                1=1
            group by
                1,
                f.code,
                f.fueltech_id
            ) as t
        group by 1, 2
        order by
            trading_day desc;
    """

    timezone = network.timezone_database
    offset = network.get_timezone(postgres_format=True)

    scada_range = get_scada_range(network=network, networks=networks_query)

    if not scada_range:
        raise Exception("Require a scada range for {}".format(network.code))

    network_region_query = ""
    fueltech_filter = ""

    if network_region:
        network_region_query = f"f.network_region='{network_region}' and "
    else:
        fueltech_filter = "and f.fueltech_id not in ('imports', 'exports')"

    network_query = "fs.network_id IN ({}) and ".format(
        networks_to_in(networks_query)
    )

    if not timezone:
        timezone = "UTC"

    if year:
        # might have to do +offset times
        year_max = "{}-12-31 23:59:59{}".format(year, offset)
        year_min = "{}-01-01 00:00:00{}".format(year, offset)

        if year == datetime.now().year:
            year_max = scada_range.get_end()

        trunc = "day"
    else:
        year_min = scada_range.get_start()
        year_max = scada_range.get_end()
        trunc = "month"

    network_region_query = ""

    if network_region:
        network_region_query = f" f.network_region='{network_region}' and"

    query = dedent(
        __query.format(
            trunc=trunc,
            year_min=year_min,
            year_max=year_max,
            timezone=timezone,
            network_query=network_query,
            network_region_query=network_region_query,
            fueltech_filter=fueltech_filter,
        )
    )

    return query



import sentry_sdk
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

from opennem.settings import settings


def setup_sentry() -> None:
    if settings.sentry_enabled and settings.debug is False:
        sentry_sdk.init(
            settings.sentry_url,
            traces_sample_rate=1.0,
            environment=settings.env,
            integrations=[RedisIntegration(), SqlalchemyIntegration()],
        )



from pathlib import Path
from typing import Optional

from pydantic import BaseSettings, PostgresDsn, RedisDsn
from pydantic.class_validators import validator
from scrapy.settings import Settings
from scrapy.utils.project import get_project_settings

SUPPORTED_LOG_LEVEL_NAMES = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class OpennemSettings(BaseSettings):
    env: str = "development"

    log_level: str = "DEBUG"

    db_url: PostgresDsn = "postgres://opennem:opennem@127.0.0.1:15433/opennem"

    cache_url: RedisDsn = "redis://127.0.0.1"

    sentry_url: Optional[str]

    prometheus_url: Optional[str]

    google_places_api_key: Optional[str] = None

    requests_cache_path: str = ".requests"

    slack_hook_url: Optional[str]

    export_local: bool = False

    s3_bucket_path: str = "s3://data.opennem.org.au/"

    interval_default: str = "15m"

    period_default: str = "7d"

    precision_default: int = 4

    # asgi server settings
    server_host: str = "0.0.0.0"
    server_port: int = 8000
    server_reload: bool = False

    # timeout on http requests
    # see opennem.utils.http
    http_timeout: int = 30

    # number of retries by default
    http_retries: int = 5

    _static_folder_path: str = "opennem/static/"

    # @todo overwrite scrapy settings here
    scrapy: Optional[Settings] = get_project_settings()

    # pylint: disable=no-self-argument
    @validator("log_level")
    def validate_log_level(cls, log_value: str) -> Optional[str]:

        _log_value = log_value.upper().strip()

        if _log_value not in SUPPORTED_LOG_LEVEL_NAMES:
            raise Exception("Invalid log level: {}".format(_log_value))

        return _log_value

    @property
    def sentry_enabled(self) -> bool:
        return self.sentry_url is not None

    @property
    def static_folder_path(self) -> str:
        _path: Path = Path(self._static_folder_path)

        if not _path.is_dir():
            raise Exception("{} is not a folder".format(_path))

        return str(_path.resolve())

    @property
    def debug(self) -> bool:
        if self.env in ["development", "staging"]:
            return True
        return False

    class Config:
        fields = {
            "env": {"env": "ENV"},
            "log_level": {"env": "LOG_LEVEL"},
            "_static_folder_path": {"env": "STATIC_PATH"},
            "export_local": {"env": "EXPORT_LOCAL"},
            "db_url": {"env": "DATABASE_HOST_URL"},
            "cache_url": {"env": "REDIS_HOST_URL"},
            "sentry_url": {"env": "SENTRY_URL"},
            "prometheus_url": {"env": "PROMETHEUS_URL"},
            "slack_hook_url": {"env": "MONITORING_SLACK_HOOK"},
            "s3_bucket_path": {"env": "S3_DATA_BUCKET_PATH"},
            "server_port": {"env": "PORT"},
            "server_host": {"env": "HOST"},
        }



"""
    settings files - read settings from env


    will read in order:

        * .env
        * .env.{environment}
        * system env
        * pydantic settings

"""
import logging
import logging.config
import os

from dotenv import load_dotenv

from opennem.settings.log import LOGGING_CONFIG
from opennem.settings.utils import load_env_file
from opennem.utils.proc import running_as_scrapy

from .schema import OpennemSettings

MODULE_DIR = os.path.dirname(__file__)

ENV = os.getenv("ENV", default="development")

env_files = load_env_file(ENV)

# Load the env files
# @TODO add logging
for _env_file in env_files:
    load_dotenv(dotenv_path=_env_file, override=True)

# @NOTE don't use pydantics env file support since it doesn't support multiple
settings: OpennemSettings = OpennemSettings()


# skip if the current cli is scrapy
if LOGGING_CONFIG and not running_as_scrapy():
    # don't mess with scrapy logging

    logging.config.dictConfig(LOGGING_CONFIG)

    log_level = logging.getLevelName(settings.log_level)

    # set root log level
    logging.root.setLevel(log_level)

    opennem_logger = logging.getLogger("opennem")
    opennem_logger.setLevel(log_level)



import logging
from logging.config import fileConfig
from typing import List

from alembic import context
from sqlalchemy import engine_from_config, pool

from opennem.db import db_connect
from opennem.db.models import opennem
from opennem.settings import settings

config = context.config
fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")

config.set_main_option("sqlalchemy.url", settings.db_url)


target_metadata = opennem.metadata


def exclude_tables_from_config(config_: str) -> List[str]:
    tables = []
    tables_ = config_.get("tables", None)
    if tables_ is not None:
        tables = tables_.split(",")
    return tables


exclude_tables = exclude_tables_from_config(
    config.get_section("alembic:exclude")
)


def include_object(object, name, type_, reflected, compare_to):
    if type_ == "table" and name in exclude_tables:
        return False
    else:
        return True


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        include_object=include_object,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # connectable = engine_from_config(
    #     config.get_section(config.config_ini_section),
    #     prefix="sqlalchemy.",
    #     # poolclass=pool.Pool(),
    # )

    engine = db_connect(timeout=600)

    with engine.connect() as connection:
        context.configure(
            include_object=include_object,
            connection=connection,
            target_metadata=target_metadata,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import logging
from typing import Dict, List

from opennem.core.dispatch_type import DispatchType
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility, Location, Station

logger = logging.getLogger(__name__)

ROOFTOP_CODE = "ROOFTOP"


STATE_NETWORK_REGION_MAP = [
    # APVI
    {"state": "NSW", "network": "APVI", "network_region": "NSW1"},
    {"state": "QLD", "network": "APVI", "network_region": "QLD1"},
    {"state": "VIC", "network": "APVI", "network_region": "VIC1"},
    {"state": "TAS", "network": "APVI", "network_region": "TAS1"},
    {"state": "SA", "network": "APVI", "network_region": "SA1"},
    {"state": "NT", "network": "APVI", "network_region": "NT1"},
    {"state": "WA", "network": "APVI", "network_region": "WEM"},
    # AEMO
    {"state": "NSW", "network": "NEM", "network_region": "NSW1"},
    {"state": "QLD", "network": "NEM", "network_region": "QLD1"},
    {"state": "VIC", "network": "NEM", "network_region": "VIC1"},
    {"state": "TAS", "network": "NEM", "network_region": "TAS1"},
    {"state": "SA", "network": "NEM", "network_region": "SA1"},
    {"state": "NT", "network": "NEM", "network_region": "NT1"},
]


def rooftop_facilities():

    session = SessionLocal()

    for state_map in STATE_NETWORK_REGION_MAP:
        state_rooftop_code = "{}_{}_{}".format(
            ROOFTOP_CODE,
            state_map["network"].upper(),
            state_map["state"].upper(),
        )

        rooftop_station = (
            session.query(Station)
            .filter_by(code=state_rooftop_code)
            .one_or_none()
        )

        if not rooftop_station:
            logger.info("Creating new station {}".format(state_rooftop_code))
            rooftop_station = Station(
                code=state_rooftop_code,
            )

        rooftop_station.name = "Rooftop Solar {}".format(state_map["state"])
        rooftop_station.description = "Solar rooftop facilities for {}".format(
            state_map["state"]
        )
        rooftop_station.approved = True
        rooftop_station.approved_by = "opennem.importer.rooftop"
        rooftop_station.created_by = "opennem.importer.rooftop"

        if not rooftop_station.location:
            rooftop_station.location = Location(state=state_map["state"])

        rooftop_fac = (
            session.query(Facility)
            .filter_by(code=state_rooftop_code)
            .one_or_none()
        )

        if not rooftop_fac:
            logger.info("Creating new facility {}".format(state_rooftop_code))
            rooftop_fac = Facility(code=state_rooftop_code)

        rooftop_fac.network_id = state_map["network"]
        rooftop_fac.network_region = state_map["network_region"]
        rooftop_fac.fueltech_id = "solar_rooftop"
        rooftop_fac.status_id = "operating"
        rooftop_fac.active = True
        rooftop_fac.dispatch_type = DispatchType.GENERATOR
        rooftop_fac.approved_by = "opennem.importer.rooftop"
        rooftop_fac.created_by = "opennem.importer.rooftop"

        rooftop_station.facilities.append(rooftop_fac)
        session.add(rooftop_fac)

        session.add(rooftop_station)

    session.commit()


AEMO_REGION_REMAP = {
    "TAS": "TAS1",
    "QLD": "QLD1",
}


def remap_aemo_region(region_name: str) -> str:
    for _region_prefix, region_code in AEMO_REGION_REMAP.items():
        if region_name.startswith(_region_prefix):
            return region_code

    return region_name


def rooftop_remap_regionids(records: List[Dict]) -> List[Dict]:
    for _rec in records:
        fac_code = _rec["facility_code"]

        rec_region_code = remap_aemo_region(fac_code)

        rooftop_fac_code = "{}_{}_{}".format(
            ROOFTOP_CODE, "NEM", rec_region_code.rstrip("1")
        )
        _rec["facility_code"] = rooftop_fac_code

    return records


if __name__ == "__main__":
    rooftop_facilities()



from datetime import timezone as timezone_native
from typing import List, Optional, Union

from pydantic import Field
from pytz import timezone as pytz_timezone

from opennem.api.time import human_to_interval
from opennem.core.time import get_interval_by_size
from opennem.schema.time import TimeInterval
from opennem.utils.timezone import get_current_timezone, get_fixed_timezone

from .core import BaseConfig


class NetworkNetworkRegion(BaseConfig):
    code: str


class NetworkSchema(BaseConfig):
    code: str
    country: str
    label: str

    regions: Optional[List[NetworkNetworkRegion]]
    timezone: Optional[str] = Field(None, description="Network timezone")
    timezone_database: Optional[str] = Field(
        "UTC", description="Database timezone format"
    )
    offset: Optional[int] = Field(
        None, description="Network time offset in minutes"
    )
    interval_size: int = Field(
        ..., description="Size of network interval in minutes"
    )

    def get_interval(self) -> Optional[TimeInterval]:
        if not self.interval_size:
            return None

        interval = get_interval_by_size(self.interval_size)

        return interval

    def get_timezone(
        self, postgres_format=False
    ) -> Union[timezone_native, pytz_timezone]:
        tz = get_current_timezone()

        if self.offset:
            tz = get_fixed_timezone(self.offset)
        if self.timezone:
            tz = pytz_timezone(self.timezone)

        if postgres_format:
            tz = str(tz)[:3]

        return tz

    @property
    def intervals_per_hour(self):
        return 60 / self.interval_size


class NetworkRegion(BaseConfig):
    code: str
    network: NetworkSchema

    timezone: Optional[str] = Field(None, description="Network timezone")
    timezone_database: Optional[str] = Field(
        "UTC", description="Database timezone format"
    )
    offset: Optional[int] = Field(
        None, description="Network time offset in minutes"
    )


# @TODO move this to db + fixture

NetworkNEM = NetworkSchema(
    code="NEM",
    label="NEM",
    country="au",
    timezone_database="AEST",
    offset=600,
    interval_size=5,
)

NetworkWEM = NetworkSchema(
    code="WEM",
    label="WEM",
    country="au",
    timezone_database="AWST",
    offset=480,
    interval_size=30,
)

NetworkAPVI = NetworkSchema(
    code="APVI",
    label="APVI",
    country="au",
    timezone_database="AEST",
    offset=600,
    interval_size=15,
)

NetworkAU = NetworkSchema(
    code="AU",
    label="AU",
    country="au",
    timezone_database="AEST",
    offset=600,
    interval_size=30,
)

NETWORKS = [NetworkNEM, NetworkWEM, NetworkAPVI, NetworkAU]



import logging
import os
from io import BytesIO

import requests
from PIL import Image
from smart_open import open

logger = logging.getLogger(__name__)

S3_EXPORT_DEFAULT_BUCKET = "s3://photos.opennem.org.au/"


UPLOAD_ARGS = {
    "ContentType": "image/jpeg",
}


def write_photo_to_s3(file_path: str, data):
    # @TODO move this to aws.py
    s3_save_path = os.path.join(S3_EXPORT_DEFAULT_BUCKET, file_path)
    write_count = 0

    with open(
        s3_save_path,
        "wb",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        write_count = fh.write(data)

    logger.info("Wrote {} to {}".format(len(data), s3_save_path))

    return write_count


def photos_process():
    """"""
    pass
    # session = SessionLocal()

    # photos: List[Photo] = session.query(Photo).filter_by(processed=False).all()

    # for photo in photos:
    #     r = write_photo_to_s3(photo.name, photo.url)


def store_photo(photo_url: str):
    img = Image.open(requests.get(photo_url, stream=True).raw)

    save_bugger = BytesIO()
    img.save(save_bugger, format="JPEG")

    write_photo_to_s3("test.jpeg", data=save_bugger.getbuffer())


if __name__ == "__main__":
    photos_process()



"""
    Loads database fixtures that are required.

"""

import logging
from datetime import datetime
from decimal import Decimal

from opennem.core.fueltechs import lookup_fueltech
from opennem.core.loader import load_data
from opennem.core.normalizers import normalize_duid, station_name_cleaner
from opennem.db import SessionLocal
from opennem.db.models.opennem import (
    BomStation,
    Facility,
    FacilityStatus,
    FuelTech,
    Network,
    NetworkRegion,
    Station,
)
from opennem.importer.compat import map_compat_facility_state

logger = logging.getLogger(__name__)


def load_fueltechs():
    """
    Load the fueltechs fixture
    """
    fixture = load_data("fueltechs.json", from_fixture=True)

    s = SessionLocal()

    for ft in fixture:
        fueltech = s.query(FuelTech).filter_by(code=ft["code"]).one_or_none()

        if not fueltech:
            fueltech = FuelTech(
                code=ft["code"],
            )

        fueltech.label = ft["label"]
        fueltech.renewable = ft["renewable"]

        try:
            s.add(fueltech)
            s.commit()
        except Exception:
            logger.error("Have {}".format(fueltech.code))


def load_facilitystatus():
    """
    Load the facility status fixture
    """
    fixture = load_data("facility_status.json", from_fixture=True)

    s = SessionLocal()

    for status in fixture:
        facility_status = (
            s.query(FacilityStatus)
            .filter_by(code=status["code"])
            .one_or_none()
        )

        if not facility_status:
            facility_status = FacilityStatus(
                code=status["code"],
            )

        facility_status.label = status["label"]

        try:
            s.add(facility_status)
            s.commit()
        except Exception:
            logger.error("Have {}".format(facility_status.code))


def load_networks():
    """
    Load the networks fixture
    """
    fixture = load_data("networks.json", from_fixture=True)

    s = SessionLocal()

    for network in fixture:
        network_model = (
            s.query(Network).filter_by(code=network["code"]).one_or_none()
        )

        if not network_model:
            network_model = Network(code=network["code"])

        network_model.label = network["label"]
        network_model.country = network["country"]
        network_model.timezone = network["timezone"]
        network_model.timezone_database = network["timezone_database"]
        network_model.offset = network["offset"]
        network_model.interval_size = network["interval_size"]

        if "export_set" in network:
            network_model.export_set = network["export_set"]

        try:
            s.add(network_model)
            s.commit()
        except Exception:
            logger.error("Have {}".format(network_model.code))


def load_network_regions():
    """
    Load the network region fixture
    """
    fixture = load_data("network_regions.json", from_fixture=True)

    s = SessionLocal()

    for network_region in fixture:
        network_region_model = (
            s.query(NetworkRegion)
            .filter_by(code=network_region["code"])
            .one_or_none()
        )

        if not network_region_model:
            network_region_model = NetworkRegion(code=network_region["code"])

        network_region_model.network_id = network_region["network_id"]

        try:
            s.add(network_region_model)
            s.commit()
        except Exception:
            logger.error("Have {}".format(network_region_model.code))


"""
    BOM Fixtures
"""


def parse_date(date_str):
    dt = datetime.strptime(date_str, "%Y%m%d")
    return dt.date()


def parse_fixed_line(line):
    """
    Parses a fixed-width CSV from the funky BOM format
    """
    return (
        str(line[:6]),
        str(line[8:11]).strip(),
        str(line[18:59]).strip(),
        parse_date(line[59:67]),
        Decimal(line[75:83]),
        Decimal(line[84:]),
    )


def load_bom_stations_csv():
    """
    Imports the BOM fixed-width stations format

    Made redundant with the new JSON
    """
    s = SessionLocal()

    station_csv = load_data("stations_db.txt", from_fixture=True)

    lines = station_csv.split("\n")

    for line in lines:
        code, state, name, registered, lng, lat = parse_fixed_line(line)

        station = s.query(BomStation).filter_by(code=code).one_or_none()

        if not station:
            station = BomStation(
                code=code,
                state=state,
                name=name,
                registered=registered,
            )

        station.geom = "SRID=4326;POINT({} {})".format(lat, lng)

        try:
            s.add(station)
            s.commit()
        except Exception:
            logger.error("Have {}".format(station.code))


def load_bom_stations_json():
    """
    Imports BOM stations into the database from bom_stations.json

    The json is obtained using scripts/bom_stations.py
    """
    session = SessionLocal()

    bom_stations = load_data("bom_stations.json", from_project=True)
    bom_capitals = load_data("bom_capitals.json", from_project=True)

    codes = []

    for bom_station in bom_stations:

        if not "code" in bom_station:
            logger.error("Invalida bom station ..")
            continue

        if bom_station["code"] in codes:
            continue

        codes.append(bom_station["code"])

        station = (
            session.query(BomStation)
            .filter_by(code=bom_station["code"])
            .one_or_none()
        )

        if not station:
            logger.info("New station: %s", bom_station["name"])

            station = BomStation(
                code=bom_station["code"],
            )

        station.name = bom_station["name_full"]
        station.name_alias = bom_station["name"]
        station.website_url = bom_station["url"]
        station.feed_url = bom_station["json_feed"]
        station.priority = 5
        station.state = bom_station["state"]
        station.altitude = bom_station["altitude"]

        if bom_station["code"] in bom_capitals:
            station.is_capital = True
            station.priority = 1

        station.geom = "SRID=4326;POINT({} {})".format(
            bom_station["lng"], bom_station["lat"]
        )

        session.add(station)

    session.commit()


def update_existing_geos():
    """
    Old method to update geos from existing facilities file on OpenNEM
    """

    station_fixture = load_data("facility_registry.json", from_fixture=True)

    stations = [{"station_code": k, **v} for k, v in station_fixture.items()]

    s = SessionLocal()

    for station_data in stations:
        station = None

        station_name = station_name_cleaner(station_data["display_name"])
        station_code = normalize_duid(station_data["station_code"])
        station_state = map_compat_facility_state(
            station_data["status"]["state"]
        )

        station = (
            s.query(Station)
            .filter(Station.network_code == station_code)
            .one_or_none()
        )

        if not station:
            logger.info("Could not find station {}".format(station_code))
            continue

        if (
            "location" in station_data
            and "latitude" in station_data["location"]
            and station_data["location"]["latitude"]
        ):
            station.geom = (
                "SRID=4326;POINT({} {})".format(
                    station_data["location"]["latitude"],
                    station_data["location"]["longitude"],
                ),
            )
            station.geocode_processed_at = datetime.now()
            station.geocode_by = "opennem"
            station.geocode_approved = True

            station.updated_by = "fixture.registry"

        s.add(station)

        logger.info(
            "Updated station geo location {} ({})".format(
                station.code,
                station.name,
            )
        )

        facilities = [
            {"code": k, **v} for k, v in stations[0]["duid_data"].items()
        ]

        # update fueltechs
        for facility_data in facilities:
            facility_duid = facility_data["code"]
            facility_fueltech = lookup_fueltech(facility_data["fuel_tech"])

            facility = (
                s.query(Facility)
                .filter(Facility.network_code == facility_duid)
                .first()
            )

            if not facility:
                logger.error(
                    "Could not find existing facility {} for station {}".format(
                        facility_duid, station_code
                    )
                )
                continue

            if not facility.fueltech_id:
                facility.fueltech_id = facility_fueltech

            if facility.fueltech_id != facility_fueltech:
                logger.error(
                    "Fueltech mismatch for {}. Old is {} and new is {}".format(
                        station_code, facility_fueltech, station.fueltech_id
                    )
                )

            s.add(facility)

        s.commit()


def load_fixtures():
    load_fueltechs()
    load_facilitystatus()
    load_networks()
    load_network_regions()
    load_bom_stations_json()


if __name__ == "__main__":
    load_fixtures()



import logging
from datetime import datetime
from io import StringIO
from typing import Any, List, Optional, Union

# from sqlalchemy.exc import StatementError
from sqlalchemy.sql.schema import Column, Table

from opennem.db import get_database_engine
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)

BULK_INSERT_QUERY = """
    CREATE TEMP TABLE __tmp_{table_name}_{tmp_table_name}
    (LIKE {table_name} INCLUDING DEFAULTS)
    ON COMMIT DROP;

    COPY __tmp_{table_name}_{tmp_table_name} FROM STDIN WITH (FORMAT CSV, HEADER TRUE, DELIMITER ',');

    INSERT INTO {table_name}
        SELECT *
        FROM __tmp_{table_name}_{tmp_table_name}
    ON CONFLICT {on_conflict}
"""

BULK_INSERT_CONFLICT_UPDATE = """
    ({pk_columns}) DO UPDATE set {update_values}
"""


def build_insert_query(
    table: Table,
    update_cols: List[Union[str, Column]] = None,
) -> str:
    """
    Builds the bulk insert query
    """
    on_conflict = "DO NOTHING"

    def get_column_name(column: Union[str, Column]) -> str:
        if isinstance(column, Column) and hasattr(column, "name"):
            return column.name
        if isinstance(column, str):
            return column.strip()
        return ""

    update_col_names = []

    if update_cols:
        update_col_names = [get_column_name(c) for c in update_cols]

    update_col_names = list(filter(lambda c: c, update_col_names))

    primary_key_columns = [
        c.name for c in table.__table__.primary_key.columns.values()
    ]

    if len(update_col_names):
        on_conflict = BULK_INSERT_CONFLICT_UPDATE.format(
            pk_columns=",".join(primary_key_columns),
            update_values=", ".join(
                [f"{n} = EXCLUDED.{n}" for n in update_col_names]
            ),
        )

    query = BULK_INSERT_QUERY.format(
        table_name=table.__table__.name,
        on_conflict=on_conflict,
        tmp_table_name=datetime.strftime(datetime.now(), "%Y%m%d%H%M%S"),
    )

    logger.debug(query)

    return query


class BulkInsertPipeline(object):
    @check_spider_pipeline
    def process_item(self, item: List[dict], spider):
        num_records = 0
        conn = get_database_engine().raw_connection()

        if not isinstance(item, list):
            spider_name = "unknown"

            if spider and hasattr(spider, "name"):
                spider_name = spider.name

            logger.error(
                "Did not get a list of items in bulk inserter from spider {}".format(
                    spider_name
                )
            )

            return {"num_records": "ERROR"}

        for single_item in item:
            if "csv" not in single_item:
                logger.error("No csv record passed to bulk inserter")
                return 0

            csv_content: StringIO = single_item["csv"]

            if "table_schema" not in single_item:
                logger.error("No table model passed to bulk inserter")
                return item

            table: Table = single_item["table_schema"]

            update_fields: Optional[List[Union[str, Column[Any]]]] = None

            if "update_fields" in single_item:
                update_fields = single_item["update_fields"]

            sql_query = build_insert_query(table, update_fields)

            try:
                cursor = conn.cursor()
                cursor.copy_expert(sql_query, csv_content)
                conn.commit()
            except Exception as generic_error:
                if hasattr(generic_error, "hide_parameters"):
                    generic_error.hide_parameters = True
                logger.error(generic_error)

            try:
                num_records += len(csv_content.getvalue().split("\n")) - 1
            except Exception:
                pass

        return {"num_records": num_records}



"""
    Create stations on NEM for interconnectors
    which provide stats on region flows.


"""

import logging
from typing import List

from opennem.core.dispatch_type import DispatchType
from opennem.core.loader import load_data
from opennem.core.networks import network_from_state, state_from_network_region
from opennem.core.parsers.aemo import (
    AEMOParserException,
    AEMOTableSchema,
    parse_aemo_csv,
)
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility, Location, Station
from opennem.schema.aemo.mms import MarketConfigInterconnector
from opennem.utils.mime import decode_bytes

logger = logging.getLogger(__name__)


def import_nem_interconnects() -> None:
    session = SessionLocal()

    # Load the MMS CSV file that contains interconnector info
    csv_data = load_data(
        "mms/PUBLIC_DVD_INTERCONNECTOR_202006010000.CSV",
        from_project=True,
    )

    # gotta be a string otherwise decode
    if not isinstance(csv_data, str):
        csv_data = decode_bytes(csv_data)

    # parse the AEMO CSV into schemas
    aemo_table_set = None

    try:
        aemo_table_set = parse_aemo_csv(csv_data)
    except AEMOParserException as e:
        logger.error(e)
        return None

    records: List[MarketConfigInterconnector] = aemo_table_set.get_table(
        "MARKET_CONFIG_INTERCONNECTOR"
    ).get_records()

    for interconnector in records:
        if not isinstance(interconnector, MarketConfigInterconnector):
            raise Exception("Not what we're looking for ")

        # skip SNOWY
        # @TODO do these need to be remapped for historical
        if interconnector.regionfrom in [
            "SNOWY1"
        ] or interconnector.regionto in ["SNOWY1"]:
            continue

        logger.debug(interconnector)

        interconnector_station = (
            session.query(Station)
            .filter_by(code=interconnector.interconnectorid)
            .filter_by(network_code=interconnector.interconnectorid)
            .one_or_none()
        )

        if not interconnector_station:
            interconnector_station = Station(
                code=interconnector.interconnectorid,
                network_code=interconnector.interconnectorid,
            )

        interconnector_station.approved = True
        interconnector_station.approved_by = __name__
        interconnector_station.created_by = __name__

        if not interconnector_station.location:
            interconnector_station.location = Location(
                state=state_from_network_region(interconnector.regionfrom)
            )

        interconnector_station.name = interconnector.description

        # for network_region in [interconnector.regionfrom, interconnector.regionto]:
        # Fac1
        int_facility = (
            session.query(Facility)
            .filter_by(code=interconnector.interconnectorid)
            .filter_by(dispatch_type=DispatchType.GENERATOR)
            .filter_by(network_id="NEM")
            .filter_by(network_region=interconnector.regionfrom)
            .one_or_none()
        )

        if not int_facility:
            int_facility = Facility(  # type: ignore
                code=interconnector.interconnectorid,
                dispatch_type=DispatchType.GENERATOR,
                network_id="NEM",
                network_region=interconnector.regionfrom,
            )

        int_facility.status_id = "operating"
        int_facility.approved = True
        int_facility.approved_by = "opennem.importer.interconnectors"
        int_facility.created_by = "opennem.importer.interconnectors"
        int_facility.fueltech_id = "exports"

        # logger.debug(
        #     "Created facility: {}".format(interconnector.interconnectorid)
        # )

        # Fac2
        _code = "{}-2".format(interconnector.interconnectorid)
        int_facility2 = (
            session.query(Facility)
            .filter_by(code=_code)
            .filter_by(dispatch_type=DispatchType.LOAD)
            .filter_by(network_region=interconnector.regionto)
            .filter_by(network_id="NEM")
            .one_or_none()
        )

        if not int_facility2:
            int_facility2 = Facility(  # type: ignore
                code=_code,
                dispatch_type=DispatchType.LOAD,
                network_id="NEM",
                network_region=interconnector.regionto,
            )

        int_facility2.status_id = "operating"
        int_facility2.approved = True
        int_facility2.approved_by = "opennem.importer.interconnectors"
        int_facility2.created_by = "opennem.importer.interconnectors"
        int_facility2.fueltech_id = "imports"

        interconnector_station.facilities.append(int_facility)
        interconnector_station.facilities.append(int_facility2)

        session.add(interconnector_station)

        logger.debug("Created station: {}".format(interconnector_station.code))

    session.commit()

    return None


if __name__ == "__main__":
    import_nem_interconnects()



import urllib
from urllib.parse import urljoin, urlparse

# Support S3 URI's in urllib
urllib.parse.uses_netloc.append("s3")
urllib.parse.uses_relative.append("s3")


def bucket_to_website(bucket_path: str, to_scheme: str = "https") -> str:
    """
    Converts a bucket path to a website path
    """
    bucket_path_parsed = urlparse(bucket_path)
    bucket_path_parsed = bucket_path_parsed._replace(scheme=to_scheme)
    return bucket_path_parsed.geturl()



import re
from datetime import datetime, timedelta
from typing import List

from pytz import timezone

UNITS = {
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


def human_to_timedelta(s: str) -> timedelta:
    return timedelta(
        **{
            UNITS.get(m.group("unit").lower(), "seconds"): int(m.group("val"))
            for m in re.finditer(
                r"(?P<val>\d+)(?P<unit>[smhdw]?)", s, flags=re.I
            )
        }
    )


def dt_floor(dt: datetime, interval=5) -> datetime:
    """
        Round a datetime down to the nearest interval in minutes

    """
    return dt - timedelta(
        minutes=dt.minute % interval,
        seconds=dt.second,
        microseconds=dt.microsecond,
    )


def dt_series(tz: timezone = timezone("UTC")) -> List[datetime]:
    """
        Generate a datetime series

    """
    end = dt_floor(tz.localize(datetime.now()))

    start = end - timedelta(days=7)
    interval = timedelta(minutes=5)

    dc_current = start
    series = []

    while dc_current <= end:
        series.append(dc_current)
        dc_current += interval

    return series



import csv
import logging
from datetime import datetime

from opennem.pipelines.nem.opennem import unit_scada_generate_facility_scada
from opennem.schema.network import NetworkWEM
from opennem.settings.scrapy import USER_AGENT
from opennem.utils.http import http

logger = logging.getLogger(__name__)

LIVE_FACILITIES = "https://aemo.com.au/aemo/data/wa/infographic/facility-intervals-last96.csv"

REQ_HEADERS = {
    "User-Agent": USER_AGENT,
}


def get_aemo_wem_live_facility_intervals_recent_date() -> datetime:
    req = http.get(LIVE_FACILITIES, headers=REQ_HEADERS)

    if req.status_code != 200:
        logger.error(
            "WEM live facility intervals returning non 200: {} {}".format(
                LIVE_FACILITIES, req.status_code
            )
        )

    csv_content = req.content
    csvreader = csv.DictReader(csv_content.decode("utf-8").split("\n"))

    if not csvreader.fieldnames or len(csvreader.fieldnames) < 1:
        logger.error(
            "WEM live facility intervals returning bad CSV: {}".format(
                LIVE_FACILITIES
            )
        )

    records = unit_scada_generate_facility_scada(
        records=csvreader,
        interval_field="PERIOD",
        facility_code_field="FACILITY_CODE",
        power_field="ACTUAL_MW",
        network=NetworkWEM,
    )

    max_date = max([i["trading_interval"] for i in records])

    return max_date



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-member
from huey import RedisHuey, crontab

from opennem.api.export.map import PriorityType
from opennem.api.export.tasks import (
    export_energy,
    export_metadata,
    export_power,
)
from opennem.exporter.geojson import export_facility_geojson
from opennem.monitors.aemo_intervals import aemo_wem_live_interval
from opennem.monitors.opennem import check_opennem_interval_delays
from opennem.monitors.opennem_metadata import check_metadata_status
from opennem.settings import settings

redis_host = None

if settings.cache_url:
    redis_host = settings.cache_url.host

huey = RedisHuey("opennem.exporter", host=redis_host)
# export tasks


@huey.periodic_task(crontab(minute="*/2"))
@huey.lock_task("schedule_live_tasks")
def schedule_live_tasks() -> None:
    export_power(priority=PriorityType.live)


@huey.periodic_task(crontab(hour="*/2"))
@huey.lock_task("schedule_power_weeklies")
def schedule_power_weeklies() -> None:
    """
    Run weekly power outputs
    """
    export_power(priority=PriorityType.history, latest=True)


@huey.periodic_task(crontab(hour="*/3"))
@huey.lock_task("schedule_power_weeklies_archive")
def schedule_power_weeklies_archive() -> None:
    """
    Run weekly power outputs entire archive
    """
    export_power(priority=PriorityType.history)


@huey.periodic_task(crontab(hour="*/1"))
@huey.lock_task("schedule_hourly_tasks")
def schedule_hourly_tasks() -> None:
    export_energy(priority=PriorityType.daily, latest=True)


@huey.periodic_task(crontab(hour="*/2"))
@huey.lock_task("schedule_daily_tasks")
def schedule_daily_tasks() -> None:
    export_energy(priority=PriorityType.daily)


@huey.periodic_task(crontab(hour="*/12"))
@huey.lock_task("schedule_energy_monthlies")
def schedule_energy_monthlies() -> None:
    export_energy(priority=PriorityType.monthly)


# geojson maps
@huey.periodic_task(crontab(minute="*/30"))
@huey.lock_task("schedule_export_geojson")
def schedule_export_geojson() -> None:
    export_facility_geojson()


# metadata
@huey.periodic_task(crontab(minute="*/30"))
@huey.lock_task("schedule_export_metadata")
def schedule_export_metadata() -> None:
    export_metadata()


# monitoring tasks
@huey.periodic_task(crontab(minute="*/15"))
@huey.lock_task("monitor_opennem_intervals")
def monitor_opennem_intervals() -> None:
    for network_code in ["NEM", "WEM"]:
        check_opennem_interval_delays(network_code)


@huey.periodic_task(crontab(minute="*/15"))
@huey.lock_task("monitor_wem_interval")
def monitor_wem_interval() -> None:
    aemo_wem_live_interval()


@huey.periodic_task(crontab(hour="*/1"))
@huey.lock_task("monitor_metadata_status")
def monitor_metadata_status() -> None:
    check_metadata_status()



from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.wem.facility_scada import (
    WemStoreFacilityIntervals,
    WemStoreFacilityScada,
    WemStoreLiveFacilityScada,
)
from opennem.spiders.wem import WemCurrentSpider, WemHistoricSpider


class WemLiveFacilityScada(WemCurrentSpider):
    # name = "au.wem.live.facility_scada"
    pipelines_extra = set([WemStoreLiveFacilityScada])

    start_url = "https://aemo.com.au/aemo/data/wa/infographic/generation.csv"


class WemLiveFacilityIntervals(WemCurrentSpider):
    name = "au.wem.live.facility_intervals"
    pipelines_extra = set(
        [WemStoreFacilityIntervals, RecordsToCSVPipeline, BulkInsertPipeline]
    )

    start_url = "https://aemo.com.au/aemo/data/wa/infographic/facility-intervals-last96.csv"


class WemCurrentFacilityScada(WemCurrentSpider):
    name = "au.wem.current.facility_scada"
    pipelines_extra = set(
        [WemStoreFacilityScada, RecordsToCSVPipeline, BulkInsertPipeline]
    )

    start_url = "http://data.wa.aemo.com.au/public/public-data/datafiles/facility-scada/facility-scada-{year}-{month}.csv"


class WemHistoricFacilityScada(WemHistoricSpider):
    name = "au.wem.historic.facility_scada"
    start_url = "http://data.wa.aemo.com.au/datafiles/facility-scada/"

    pipelines_extra = set(
        [WemStoreFacilityScada, RecordsToCSVPipeline, BulkInsertPipeline]
    )



"""
Tasks file to expor JSON's to S3 or locally for the
opennem website

This is the most frequently accessed content that doesn't
require the API


"""

import logging
from datetime import datetime
from typing import List, Optional

from opennem.api.export.controllers import (
    energy_fueltech_daily,
    power_week,
    weather_daily,
)
from opennem.api.export.map import (
    PriorityType,
    StatExport,
    StatType,
    get_export_map,
)
from opennem.api.export.utils import write_output
from opennem.settings import settings

logger = logging.getLogger(__name__)


def export_power(
    stats: List[StatExport] = None,
    priority: Optional[PriorityType] = None,
    latest: Optional[bool] = False,
) -> None:
    """
    Export power stats from the export map


    """
    if not stats:
        export_map = get_export_map()
        stats = export_map.get_by_stat_type(StatType.power, priority)

    output_count: int = 0

    for power_stat in stats:
        if power_stat.stat_type != StatType.power:
            continue

        if output_count >= 1 and latest:
            return None

        stat_set = power_week(
            date_range=power_stat.date_range,
            period=power_stat.period,
            network_code=power_stat.network.code,
            network_region_code=power_stat.network_region,
            networks_query=power_stat.networks,
        )

        if not stat_set:
            logger.info(
                "No power stat set for {} {} {}".format(
                    power_stat.period,
                    power_stat.networks,
                    power_stat.network_region,
                )
            )
            continue

        if power_stat.bom_station:
            weather_set = weather_daily(
                station_code=power_stat.bom_station,
                network_code=power_stat.network.code,
                include_min_max=False,
                period_human="7d",
                unit_name="temperature",
            )

            stat_set.append_set(weather_set)

        write_output(power_stat.path, stat_set)
        output_count += 1


def export_energy(
    stats: List[StatExport] = None,
    priority: Optional[PriorityType] = None,
    latest: Optional[bool] = False,
) -> None:
    """
    Export energy stats from the export map


    """
    if not stats:
        export_map = get_export_map()
        stats = export_map.get_by_stat_type(StatType.energy, priority)

    CURRENT_YEAR = datetime.now().year

    for energy_stat in stats:
        if energy_stat.stat_type != StatType.energy:
            continue

        if energy_stat.year:

            if latest and energy_stat.year != CURRENT_YEAR:
                continue

            stat_set = energy_fueltech_daily(
                year=energy_stat.year,
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if not stat_set:
                continue

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)

        elif energy_stat.period and energy_stat.period.period_human == "all":
            stat_set = energy_fueltech_daily(
                interval_size="1M",
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if not stat_set:
                continue

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)


def export_metadata() -> bool:
    """
    Export metadata


    """
    _export_map_out = get_export_map()

    # this is a hack because pydantic doesn't
    # serialize properties
    for r in _export_map_out.resources:
        r.file_path = r.path

    wrote_bytes = write_output("metadata.json", _export_map_out)

    if wrote_bytes and wrote_bytes > 0:
        return True

    return False


if __name__ == "__main__":
    if settings.env in ["development", "staging"]:
        export_power(priority=PriorityType.live)
        export_energy(latest=True)
        export_metadata()
        export_power(priority=PriorityType.history)
    else:
        export_power(priority=PriorityType.live)
        export_energy(latest=True)
        export_energy()
        export_metadata()
        export_power(priority=PriorityType.history)
